import json
from datetime import timedelta

from django.db import models
from django.utils import timezone

from channels import DEFAULT_CHANNEL_LAYER, Channel, channel_layers


class DelayedMessageQuerySet(models.QuerySet):

    def is_due(self):
        return self.filter(due_date__lte=timezone.now())


class DelayedMessage(models.Model):

    due_date = models.DateTimeField(db_index=True)
    channel_name = models.CharField(max_length=512)
    content = models.TextField()

    objects = DelayedMessageQuerySet.as_manager()

    @property
    def delay(self):
        return self._delay

    @delay.setter
    def delay(self, milliseconds):
        self._delay = milliseconds
        self.due_date = timezone.now() + timedelta(milliseconds=milliseconds)

    def send(self, channel_layer=None, requeue_delay=1000):
        """
        Sends the message on the configured channel with the stored content.

        Deletes the DelayedMessage record if successfully sent.

        Args:
            channel_layer: optional channel_layer to use
            requeue_delay: if the channel is full, milliseconds to wait before requeue
        """
        channel_layer = channel_layer or channel_layers[DEFAULT_CHANNEL_LAYER]
        try:
            Channel(self.channel_name, channel_layer=channel_layer).send(json.loads(self.content), immediately=True)
            self.delete()
        except channel_layer.ChannelFull:
            self.delay = requeue_delay
            self.save()



import importlib
from distutils.version import StrictVersion

required_versions = {
    "asgi_rabbitmq": "0.4.0",
    "asgi_redis": "1.2.0",
    "asgi_ipc": "1.3.0",
}


def check_all():
    """
    Checks versions of all the possible packages you have installed so that
    we can easily warn people about incompatible versions.

    This is needed as there are some packages (e.g. asgi_redis) that we cannot
    declare dependencies on as they are not _required_. People usually remember
    to upgrade their Channels package so this is where we check.
    """
    for package, version in required_versions.items():
        try:
            module = importlib.import_module(package)
        except ImportError:
            continue
        else:
            if StrictVersion(version) > StrictVersion(module.__version__):
                raise RuntimeError("Your version of %s is too old - it must be at least %s" % (
                    package,
                    version,
                ))



from __future__ import unicode_literals

import six


class ConsumeLater(Exception):
    """
    Exception that says that the current message should be re-queued back
    onto its channel as it's not ready to be consumed yet (e.g. global order
    is being enforced)
    """
    pass


class ResponseLater(Exception):
    """
    Exception raised inside a Django view when the view has passed
    responsibility for the response to another consumer, and so is not
    returning a response.
    """
    pass


class RequestTimeout(Exception):
    """
    Raised when it takes too long to read a request body.
    """
    pass


class RequestAborted(Exception):
    """
    Raised when the incoming request tells us it's aborted partway through
    reading the body.
    """
    pass


class DenyConnection(Exception):
    """
    Raised during a websocket.connect (or other supported connection) handler
    to deny the connection.
    """
    pass


class ChannelSocketException(Exception):
    """
    Base Exception is intended to run some action ('run' method)
    when it is raised at a consumer body
    """

    def run(self, message):
        raise NotImplementedError


class WebsocketCloseException(ChannelSocketException):
    """
    ChannelSocketException based exceptions for close websocket connection with code
    """

    def __init__(self, code=None):
        if code is not None and not isinstance(code, six.integer_types) \
                and code != 1000 and not (3000 <= code <= 4999):
            raise ValueError("invalid close code {} (must be 1000 or from [3000, 4999])".format(code))
        self._code = code

    def run(self, message):
        if message.reply_channel.name.split('.')[0] != "websocket":
            raise ValueError("You cannot raise CloseWebsocketError from a non-websocket handler.")
        message.reply_channel.send({"close": self._code or True})


class SendNotAvailableOnDemultiplexer(Exception):
    """
    Raised when trying to send with a WebsocketDemultiplexer. Use the multiplexer instead.
    """
    pass



from __future__ import unicode_literals

import copy
import threading
import time

from .channel import Channel
from .signals import consumer_finished, consumer_started


class Message(object):
    """
    Represents a message sent over a Channel.

    The message content is a dict called .content, while
    reply_channel is an optional extra attribute representing a channel
    to use to reply to this message's end user, if that makes sense.
    """

    def __init__(self, content, channel_name, channel_layer):
        self.content = content
        self.channel = Channel(
            channel_name,
            channel_layer=channel_layer,
        )
        self.channel_layer = channel_layer
        if content.get("reply_channel", None):
            self.reply_channel = Channel(
                content["reply_channel"],
                channel_layer=self.channel_layer,
            )
        else:
            self.reply_channel = None

    def __getitem__(self, key):
        return self.content[key]

    def __setitem__(self, key, value):
        self.content[key] = value

    def __contains__(self, key):
        return key in self.content

    def keys(self):
        return self.content.keys()

    def values(self):
        return self.content.values()

    def items(self):
        return self.content.items()

    def get(self, key, default=None):
        return self.content.get(key, default)

    def copy(self):
        """
        Returns a safely content-mutable copy of this Message.
        """
        return self.__class__(
            copy.deepcopy(self.content),
            self.channel.name,
            self.channel_layer,
        )


class PendingMessageStore(object):
    """
    Singleton object used for storing pending messages that should be sent
    to a channel or group when a consumer finishes.

    Will retry when it sees ChannelFull up to a limit; if you want more control
    over this, change to `immediately=True` in your send method and handle it
    yourself.
    """

    threadlocal = threading.local()

    retry_time = 2  # seconds
    retry_interval = 0.2  # seconds

    def prepare(self, **kwargs):
        """
        Sets the message store up to receive messages.
        """
        self.threadlocal.messages = []

    @property
    def active(self):
        """
        Returns if the pending message store can be used or not
        (it can only be used inside consumers)
        """
        return hasattr(self.threadlocal, "messages")

    def append(self, sender, message):
        self.threadlocal.messages.append((sender, message))

    def send_and_flush(self, **kwargs):
        for sender, message in getattr(self.threadlocal, "messages", []):
            # Loop until the retry time limit is hit
            started = time.time()
            while time.time() - started < self.retry_time:
                try:
                    sender.send(message, immediately=True)
                except sender.channel_layer.ChannelFull:
                    time.sleep(self.retry_interval)
                    continue
                else:
                    break
            # If we didn't break out, we failed to send, so do a nice exception
            else:
                raise RuntimeError(
                    "Failed to send queued message to %s after retrying for %.2fs.\n"
                    "You need to increase the consumption rate on this channel, its capacity,\n"
                    "or handle the ChannelFull exception yourself after adding\n"
                    "immediately=True to send()." % (sender, self.retry_time)
                )
        delattr(self.threadlocal, "messages")


pending_message_store = PendingMessageStore()
consumer_started.connect(pending_message_store.prepare)
consumer_finished.connect(pending_message_store.send_and_flush)



import multiprocessing

import django
from daphne.server import Server
from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.db import connections
from django.db.utils import load_backend
from django.test.testcases import TransactionTestCase
from django.test.utils import modify_settings, override_settings
from twisted.internet import reactor

from .. import DEFAULT_CHANNEL_LAYER
from ..asgi import ChannelLayerManager
from ..staticfiles import StaticFilesConsumer
from ..worker import Worker, WorkerGroup

# NOTE: We use ChannelLayerManager to prevent layer instance sharing
# between forked process.  Some layers implementations create
# connections inside the __init__ method.  After forking child
# processes can lose the ability to use this connection and typically
# stuck on some network operation.  To prevent this we use new
# ChannelLayerManager each time we want to initiate default layer.
# This gives us guaranty that new layer instance will be created and
# new connection will be established.


class ProcessSetup(multiprocessing.Process):
    """Common initialization steps for test subprocess."""

    def common_setup(self):

        self.setup_django()
        self.setup_databases()
        self.override_settings()

    def setup_django(self):

        if django.VERSION >= (1, 10):
            django.setup(set_prefix=False)
        else:
            django.setup()

    def setup_databases(self):

        for alias, db in self.databases.items():
            backend = load_backend(db['ENGINE'])
            conn = backend.DatabaseWrapper(db, alias)
            if django.VERSION >= (1, 9):
                connections[alias].creation.set_as_test_mirror(
                    conn.settings_dict,
                )
            else:
                test_db_name = conn.settings_dict['NAME']
                connections[alias].settings_dict['NAME'] = test_db_name

    def override_settings(self):

        if self.overridden_settings:
            overridden = override_settings(**self.overridden_settings)
            overridden.enable()

        if self.modified_settings:
            modified = modify_settings(self.modified_settings)
            modified.enable()


class WorkerProcess(ProcessSetup):

    def __init__(self, is_ready, n_threads, overridden_settings,
                 modified_settings, databases, serve_static):

        self.is_ready = is_ready
        self.n_threads = n_threads
        self.overridden_settings = overridden_settings
        self.modified_settings = modified_settings
        self.databases = databases
        self.serve_static = serve_static
        super(WorkerProcess, self).__init__()
        self.daemon = True

    def run(self):

        try:
            self.common_setup()
            channel_layers = ChannelLayerManager()
            channel_layer = channel_layers.make_test_backend(DEFAULT_CHANNEL_LAYER)
            if self.serve_static and apps.is_installed('django.contrib.staticfiles'):
                channel_layer.router.check_default(http_consumer=StaticFilesConsumer())
            else:
                channel_layer.router.check_default()
            if self.n_threads == 1:
                self.worker = Worker(
                    channel_layer=channel_layer,
                    signal_handlers=False,
                )
            else:
                self.worker = WorkerGroup(
                    channel_layer=channel_layer,
                    signal_handlers=False,
                    n_threads=self.n_threads,
                )
            self.worker.ready()
            self.is_ready.set()
            self.worker.run()
        except Exception:
            self.is_ready.set()
            raise


class DaphneProcess(ProcessSetup):

    def __init__(self, host, port_storage, is_ready, overridden_settings,
                 modified_settings, databases):

        self.host = host
        self.port_storage = port_storage
        self.is_ready = is_ready
        self.overridden_settings = overridden_settings
        self.modified_settings = modified_settings
        self.databases = databases
        super(DaphneProcess, self).__init__()
        self.daemon = True

    def run(self):

        try:
            self.common_setup()
            channel_layers = ChannelLayerManager()
            channel_layer = channel_layers.make_test_backend(DEFAULT_CHANNEL_LAYER)
            self.server = Server(
                channel_layer=channel_layer,
                endpoints=['tcp:interface=%s:port=0' % (self.host)],
                signal_handlers=False,
            )
            reactor.callLater(0.5, self.resolve_port)
            self.server.run()
        except Exception:
            self.is_ready.set()
            raise

    def resolve_port(self):

        port = self.server.listeners[0].result.getHost().port
        self.port_storage.value = port
        self.is_ready.set()


class ChannelLiveServerTestCase(TransactionTestCase):
    """
    Does basically the same as TransactionTestCase but also launches a
    live Daphne server and Channels worker in a separate process, so
    that the tests may use another test framework, such as Selenium,
    instead of the built-in dummy client.
    """

    host = 'localhost'
    ProtocolServerProcess = DaphneProcess
    WorkerProcess = WorkerProcess
    worker_threads = 1
    serve_static = True

    @property
    def live_server_url(self):

        return 'http://%s:%s' % (self.host, self._port_storage.value)

    @property
    def live_server_ws_url(self):

        return 'ws://%s:%s' % (self.host, self._port_storage.value)

    def _pre_setup(self):

        for connection in connections.all():
            if self._is_in_memory_db(connection):
                raise ImproperlyConfigured(
                    'ChannelLiveServerTestCase can not be used with in memory databases'
                )

        channel_layers = ChannelLayerManager()
        if len(channel_layers.configs) > 1:
            raise ImproperlyConfigured(
                'ChannelLiveServerTestCase does not support multiple CHANNEL_LAYERS at this time'
            )

        channel_layer = channel_layers.make_test_backend(DEFAULT_CHANNEL_LAYER)
        if 'flush' in channel_layer.extensions:
            channel_layer.flush()

        super(ChannelLiveServerTestCase, self)._pre_setup()

        server_ready = multiprocessing.Event()
        self._port_storage = multiprocessing.Value('i')
        self._server_process = self.ProtocolServerProcess(
            self.host,
            self._port_storage,
            server_ready,
            self._overridden_settings,
            self._modified_settings,
            connections.databases,
        )
        self._server_process.start()
        server_ready.wait()

        worker_ready = multiprocessing.Event()
        self._worker_process = self.WorkerProcess(
            worker_ready,
            self.worker_threads,
            self._overridden_settings,
            self._modified_settings,
            connections.databases,
            self.serve_static,
        )
        self._worker_process.start()
        worker_ready.wait()

    def _post_teardown(self):

        self._server_process.terminate()
        self._server_process.join()
        self._worker_process.terminate()
        self._worker_process.join()
        super(ChannelLiveServerTestCase, self)._post_teardown()

    def _is_in_memory_db(self, connection):
        """Check if DatabaseWrapper holds in memory database."""

        if connection.vendor == 'sqlite':
            if django.VERSION >= (1, 11):
                return connection.is_in_memory_db()
            else:
                return connection.is_in_memory_db(
                    connection.settings_dict['NAME'],
                )



from django.core.serializers.json import DjangoJSONEncoder, json

from ..auth import channel_and_http_session_user_from_http, channel_session_user_from_http
from ..channel import Group
from ..exceptions import SendNotAvailableOnDemultiplexer
from ..sessions import enforce_ordering
from .base import BaseConsumer


class WebsocketConsumer(BaseConsumer):
    """
    Base WebSocket consumer. Provides a general encapsulation for the
    WebSocket handling model that other applications can build on.
    """

    # You shouldn't need to override this
    method_mapping = {
        "websocket.connect": "raw_connect",
        "websocket.receive": "raw_receive",
        "websocket.disconnect": "raw_disconnect",
    }

    # Turning this on passes the user over from the HTTP session on connect,
    # implies channel_session_user
    http_user = False
    http_user_and_session = False

    # Set to True if you want the class to enforce ordering for you
    strict_ordering = False

    groups = None

    def get_handler(self, message, **kwargs):
        """
        Pulls out the path onto an instance variable, and optionally
        adds the ordering decorator.
        """
        # HTTP user implies channel session user
        if self.http_user or self.http_user_and_session:
            self.channel_session_user = True
        # Get super-handler
        self.path = message['path']
        handler = super(WebsocketConsumer, self).get_handler(message, **kwargs)
        # Optionally apply HTTP transfer
        if self.http_user_and_session:
            handler = channel_and_http_session_user_from_http(handler)
        elif self.http_user:
            handler = channel_session_user_from_http(handler)
        # Ordering decorators
        if self.strict_ordering:
            return enforce_ordering(handler, slight=False)
        elif getattr(self, "slight_ordering", False):
            raise ValueError("Slight ordering is now always on. Please remove `slight_ordering=True`.")
        else:
            return handler

    def connection_groups(self, **kwargs):
        """
        Group(s) to make people join when they connect and leave when they
        disconnect. Make sure to return a list/tuple, not a string!
        """
        return self.groups or []

    def raw_connect(self, message, **kwargs):
        """
        Called when a WebSocket connection is opened. Base level so you don't
        need to call super() all the time.
        """
        for group in self.connection_groups(**kwargs):
            Group(group, channel_layer=message.channel_layer).add(message.reply_channel)
        self.connect(message, **kwargs)

    def connect(self, message, **kwargs):
        """
        Called when a WebSocket connection is opened.
        """
        self.message.reply_channel.send({"accept": True})

    def raw_receive(self, message, **kwargs):
        """
        Called when a WebSocket frame is received. Decodes it and passes it
        to receive().
        """
        if "text" in message:
            self.receive(text=message['text'], **kwargs)
        else:
            self.receive(bytes=message['bytes'], **kwargs)

    def receive(self, text=None, bytes=None, **kwargs):
        """
        Called with a decoded WebSocket frame.
        """
        pass

    def send(self, text=None, bytes=None, close=False):
        """
        Sends a reply back down the WebSocket
        """
        message = {}
        if close:
            message["close"] = close
        if text is not None:
            message["text"] = text
        elif bytes is not None:
            message["bytes"] = bytes
        else:
            raise ValueError("You must pass text or bytes")
        self.message.reply_channel.send(message)

    @classmethod
    def group_send(cls, name, text=None, bytes=None, close=False):
        message = {}
        if close:
            message["close"] = close
        if text is not None:
            message["text"] = text
        elif bytes is not None:
            message["bytes"] = bytes
        else:
            raise ValueError("You must pass text or bytes")
        Group(name).send(message)

    def close(self, status=True):
        """
        Closes the WebSocket from the server end
        """
        self.message.reply_channel.send({"close": status})

    def raw_disconnect(self, message, **kwargs):
        """
        Called when a WebSocket connection is closed. Base level so you don't
        need to call super() all the time.
        """
        for group in self.connection_groups(**kwargs):
            Group(group, channel_layer=message.channel_layer).discard(message.reply_channel)
        self.disconnect(message, **kwargs)

    def disconnect(self, message, **kwargs):
        """
        Called when a WebSocket connection is closed.
        """
        pass


class JsonWebsocketConsumer(WebsocketConsumer):
    """
    Variant of WebsocketConsumer that automatically JSON-encodes and decodes
    messages as they come in and go out. Expects everything to be text; will
    error on binary data.
    """

    def raw_receive(self, message, **kwargs):
        if "text" in message:
            self.receive(self.decode_json(message['text']), **kwargs)
        else:
            raise ValueError("No text section for incoming WebSocket frame!")

    def receive(self, content, **kwargs):
        """
        Called with decoded JSON content.
        """
        pass

    def send(self, content, close=False):
        """
        Encode the given content as JSON and send it to the client.
        """
        super(JsonWebsocketConsumer, self).send(text=self.encode_json(content), close=close)

    @classmethod
    def decode_json(cls, text):
        return json.loads(text)

    @classmethod
    def encode_json(cls, content):
        return json.dumps(content)

    @classmethod
    def group_send(cls, name, content, close=False):
        WebsocketConsumer.group_send(name, cls.encode_json(content), close=close)


class WebsocketMultiplexer(object):
    """
    The opposite of the demultiplexer, to send a message though a multiplexed channel.

    The multiplexer object is passed as a kwargs to the consumer when the message is dispatched.
    This pattern allows the consumer class to be independent of the stream name.
    """

    stream = None
    reply_channel = None

    def __init__(self, stream, reply_channel):
        self.stream = stream
        self.reply_channel = reply_channel

    def send(self, payload):
        """Multiplex the payload using the stream name and send it."""
        self.reply_channel.send(self.encode(self.stream, payload))

    @classmethod
    def encode_json(cls, content):
        return json.dumps(content, cls=DjangoJSONEncoder)

    @classmethod
    def encode(cls, stream, payload):
        """
        Encodes stream + payload for outbound sending.
        """
        content = {"stream": stream, "payload": payload}
        return {"text": cls.encode_json(content)}

    @classmethod
    def group_send(cls, name, stream, payload, close=False):
        message = cls.encode(stream, payload)
        if close:
            message["close"] = True
        Group(name).send(message)


class WebsocketDemultiplexer(JsonWebsocketConsumer):
    """
    JSON-understanding WebSocket consumer subclass that handles demultiplexing
    streams using a "stream" key in a top-level dict and the actual payload
    in a sub-dict called "payload". This lets you run multiple streams over
    a single WebSocket connection in a standardised way.

    Incoming messages on streams are dispatched to consumers so you can
    just tie in consumers the normal way. The reply_channels are kept so
    sessions/auth continue to work. Payloads must be a dict at the top level,
    so they fulfill the Channels message spec.

    To answer with a multiplexed message, a multiplexer object
    with "send" and "group_send" methods is forwarded to the consumer as a kwargs
    "multiplexer".

    Set a mapping of streams to consumer classes in the "consumers" keyword.
    """

    # Put your JSON consumers here: {stream_name : consumer}
    consumers = {}

    # Optionally use a custom multiplexer class
    multiplexer_class = WebsocketMultiplexer

    def receive(self, content, **kwargs):
        """Forward messages to all consumers."""
        # Check the frame looks good
        if isinstance(content, dict) and "stream" in content and "payload" in content:
            # Match it to a channel
            for stream, consumer in self.consumers.items():
                if stream == content['stream']:
                    # Extract payload and add in reply_channel
                    payload = content['payload']
                    if not isinstance(payload, dict):
                        raise ValueError("Multiplexed frame payload is not a dict")
                    # The json consumer expects serialized JSON
                    self.message.content['text'] = self.encode_json(payload)
                    # Send demultiplexer to the consumer, to be able to answer
                    kwargs['multiplexer'] = self.multiplexer_class(stream, self.message.reply_channel)
                    # Patch send to avoid sending not formatted messages from the consumer
                    if hasattr(consumer, "send"):
                        consumer.send = self.send
                    # Dispatch message
                    consumer(self.message, **kwargs)
                    return

            raise ValueError("Invalid multiplexed frame received (stream not mapped)")
        else:
            raise ValueError("Invalid multiplexed **frame received (no channel/payload key)")

    def connect(self, message, **kwargs):
        """Forward connection to all consumers."""
        self.message.reply_channel.send({"accept": True})
        for stream, consumer in self.consumers.items():
            kwargs['multiplexer'] = self.multiplexer_class(stream, self.message.reply_channel)
            consumer(message, **kwargs)

    def disconnect(self, message, **kwargs):
        """Forward disconnection to all consumers."""
        for stream, consumer in self.consumers.items():
            kwargs['multiplexer'] = self.multiplexer_class(stream, self.message.reply_channel)
            consumer(message, **kwargs)

    def send(self, *args):
        raise SendNotAvailableOnDemultiplexer("Use multiplexer.send of the multiplexer kwarg.")

    @classmethod
    def group_send(cls, name, stream, payload, close=False):
        raise SendNotAvailableOnDemultiplexer("Use WebsocketMultiplexer.group_send")



from functools import update_wrapper

from django.conf import settings
from django.http.request import validate_host

from ..exceptions import DenyConnection

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


class BaseOriginValidator(object):
    """
    Base class-based decorator for origin validation of WebSocket connect
    messages.

    This base class handles parsing of the origin header. When the origin header
    is missing, empty or contains non-ascii characters, it raises a
    DenyConnection exception to reject the connection.

    Subclasses must overwrite the method validate_origin(self, message, origin)
    to return True when a message should be accepted, False otherwise.
    """

    def __init__(self, func):
        update_wrapper(self, func)
        self.func = func

    def __call__(self, message, *args, **kwargs):
        origin = self.get_origin(message)
        if not self.validate_origin(message, origin):
            raise DenyConnection
        return self.func(message, *args, **kwargs)

    def get_header(self, message, name):
        headers = message.content['headers']
        for header in headers:
            try:
                if header[0] == name:
                    return header[1:]
            except IndexError:
                continue
        raise KeyError('No header named "{}"'.format(name))

    def get_origin(self, message):
        """
        Returns the origin of a WebSocket connect message.

        Raises DenyConnection for messages with missing or non-ascii Origin
        header.
        """
        try:
            header = self.get_header(message, b'origin')[0]
        except (IndexError, KeyError):
            raise DenyConnection
        try:
            origin = header.decode('ascii')
        except UnicodeDecodeError:
            raise DenyConnection
        return origin

    def validate_origin(self, message, origin):
        """
        Validates the origin of a WebSocket connect message.

        Must be overwritten by subclasses.
        """
        raise NotImplemented('You must overwrite this method.')


class AllowedHostsOnlyOriginValidator(BaseOriginValidator):
    """
    Class-based decorator for websocket consumers that checks that
    the origin is allowed according to the ALLOWED_HOSTS settings.
    """

    def validate_origin(self, message, origin):
        allowed_hosts = settings.ALLOWED_HOSTS
        if settings.DEBUG and not allowed_hosts:
            allowed_hosts = ['localhost', '127.0.0.1', '[::1]']

        origin_hostname = urlparse(origin).hostname
        valid = (origin_hostname and
                 validate_host(origin_hostname, allowed_hosts))
        return valid


allowed_hosts_only = AllowedHostsOnlyOriginValidator



from django.apps import AppConfig

from .binding.base import BindingMetaclass
from .package_checks import check_all


class ChannelsConfig(AppConfig):

    name = "channels"
    verbose_name = "Channels"

    def ready(self):
        # Check versions
        check_all()
        # Do django monkeypatches
        from .hacks import monkeypatch_django
        monkeypatch_django()
        # Instantiate bindings
        BindingMetaclass.register_all()



from setuptools import find_packages, setup
from channels import __version__

setup(
    name='channels',
    version=__version__,
    url='http://github.com/django/channels',
    author='Django Software Foundation',
    author_email='foundation@djangoproject.com',
    description="Brings event-driven capabilities to Django with a channel system. Django 1.8 and up only.",
    license='BSD',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    install_requires=[
        'Django>=1.8',
        'asgiref~=1.1',
        'daphne>=1.2.0',
    ],
    extras_require={
        'tests': [
            'coverage',
            'mock ; python_version < "3.0"',
            'flake8>=2.0,<3.0',
            'isort',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



from .base import TransactionChannelTestCase, ChannelTestCase, Client, apply_routes  # NOQA isort:skip
from .http import HttpClient  # NOQA isort:skip
from .websocket import WSClient  # NOQA isort:skip
from .liveserver import ChannelLiveServerTestCase  # NOQA isort:skip
